// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 25 May 2024
// License: This work is licensed under Creative Commons Attribution-ShareAlike 4.0 International 
// (CC BY-SA 4.0). https://creativecommons.org/licenses/by-sa/4.0/

#include "MFTP.h"
#include <vector>
#include <cstdint>
#include <iostream>
#include <string>
#include <chrono>
#include <thread>
#include <algorithm> // for std::min, std::max

/*
Explanation

Constructor:

Initializes the function pointers for CRC calculation, logging, and UDP communication.
These methods use the function pointers provided during initialization
to perform their respective tasks.


applyExponentialBackoff Method:

Implements an exponential backoff mechanism for handling retries with increasing time intervals.
Integration
*/


MFTP::MFTP(CRCFunction crcFunc, LoggingFunction logFunc, UDPSendFunction udpSendFunc, UDPReceiveFunction udpReceiveFunc)
    : crcFunction(crcFunc), loggingFunction(logFunc), udpSendFunction(udpSendFunc), udpReceiveFunction(udpReceiveFunc)
{}




uint32_t MFTP::calculateCRC32(char* data, uint32_t size)
{
    return crcFunction(data, size);
}

void MFTP::log(const std::string& message)
{
    loggingFunction(message);
}

void MFTP::udpSend(const std::vector<uint8_t>& data)
{
    udpSendFunction(data);
}

std::vector<uint8_t> MFTP::udpReceive(bool* pbSuccess)
{
    return udpReceiveFunction(pbSuccess);
}

void MFTP::applyExponentialBackoff(uint32_t& current_timeout, uint32_t& resend_attempts)
{
#undef min // #defined in Windows.h, hence undefining it to use std::min
    std::this_thread::sleep_for(std::chrono::milliseconds(current_timeout));
    current_timeout = std::min(current_timeout * 2, SOCKET_TIMEOUT);
    ++resend_attempts;
}
