// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 25 May 2024
// License: Creative Commons - Commercial Use

#include <iostream>
#include "Server.h"


bool Server::receiveFile(std::string& filename, std::vector<uint8_t>& file_data)
{
    bool redundancy_checking = true;
    bool bSuccess = false;
    uint32_t file_size = 0;

	std::cout << "Awaiting packets" << std::endl;
    auto recv_data = udpReceive(&bSuccess);
    if (!bSuccess)
    {
        log("udpReceive failed - socket error or time out!");
        return false;
    }
    if (recv_data.empty())return false;
	log("Received HDR_MSG with " + std::to_string(recv_data.size()) + " bytes of data.");

	file_size = *reinterpret_cast<const uint32_t*>(recv_data.data());
	filename.assign(reinterpret_cast<const char*>(recv_data.data() + sizeof(uint32_t)), recv_data.size() - sizeof(uint32_t));
	std::cout << "Received filename: " << filename << std::endl;
	std::cout << "Expected file size: " << file_size << std::endl;

	log("Received filename: " + filename);
	log("Expected file size: " + std::to_string(file_size));

	current_crc32 = calculateCRC32((char*)recv_data.data(), recv_data.size());

	log("Received CRC32: " + std::to_string(current_crc32));

    sendAckMessage(current_block_number, current_crc32);

    ZeroMemory(DataBuffer, DATA_MSG_SIZE);
    while (true)
    {
		memcpy(PreviousDataBuffer, DataBuffer, DATA_MSG_SIZE);
		ZeroMemory(DataBuffer, DATA_MSG_SIZE);
		auto recv_data = udpReceive(&bSuccess);
		if (!bSuccess)
		{
			log("Socket error or timeout!");
			break;
		}

		if (recv_data.empty())
		{
			// This an EOF_MSG. We are done...
			std::cout << std::endl;
			std::cout << "Received EOF_MSG. Received file integrity assured" << std::endl;
			Sleep(2500);
			log("Received EOF_MSG.");
			log("file_size_bytes expected: " + std::to_string(file_size));
			log("file_size_bytes_recd: " + std::to_string(file_size_bytes_recd));
			log("file_data.size(): " + std::to_string(file_data.size()));
			return true;
		}


		// Copy the block of data into the DataBuffer
		memcpy(DataBuffer, recv_data.data(), recv_data.size());
		previous_crc32 = current_crc32;
		current_crc32 = calculateCRC32((char*)DataBuffer, recv_data.size());

		// This could have been a DATA_MSG: 1 to 512 bytes (excluding a msg of SHORT_MSG_SIZE)
		// Or it could have been a SHORT_MSG: 8 bytes
		// Which could be either an REP_MSG or a DUP_MSG

		if (recv_data.size() == DATA_MSG_SIZE) // the expected case
		{
			if (redundancy_checking && current_crc32 == previous_crc32)
			{
				log("Received duplicate block " + std::to_string(current_block_number) + ", discarding.");
				// silently discard the duplicate block
			}
			else
			{
				// Append data to file_data vector
				uint32_t start_index = current_block_number * DATA_MSG_SIZE;
				if (file_data.size() < start_index + recv_data.size())
				{
					file_data.resize(start_index + recv_data.size());
				}
				std::copy(recv_data.begin(), recv_data.end(), file_data.begin() + start_index);
				log("Received DATA_MSG block " + std::to_string(current_block_number) + " with size " + std::to_string(recv_data.size()));
				file_size_bytes_recd += recv_data.size();
				sendAckMessage(current_block_number, current_crc32);
				current_block_number++;
				if (current_block_number == 1)
				{
					std::cout << "Receiving data";
				}
				if ((current_block_number % 100) == 0)
				{
					// Send to std::cout
					std::cout << ".";
				}

			}
			redundancy_checking = true;
		}
		else if (recv_data.size() == SHORT_MSG_SIZE)
		{
			// This is eith a REP_MSG, which is a CRC32 followed by a BLOCK_NUMBER
			// Or it is a DUP_MSG, which is a CRC32 followed by a DUP_WARN (0xFFFFFFFF)
			// Which is it?
			uint32_t* crc32_ptr = reinterpret_cast<uint32_t*>(DataBuffer);
			uint32_t* blocknum_ptr = reinterpret_cast<uint32_t*>(DataBuffer + sizeof(uint32_t));
			if (*blocknum_ptr == DUP_WARN)
			{
				// This is a DUP_MSG
				// The next message will be a DATA_MSG with the same CRC32
				log("Received DUP_WARN_MSG for block " + std::to_string(current_block_number) + ", disabling redundancy checking for next block.");
				redundancy_checking = false;
			}
			else
			{
				// This is a REP_MSG - meaning current data is same as previous data
				// so of course the data was not resent...
				// Copy previous buffer to file_data vector
				uint32_t start_index = current_block_number * DATA_MSG_SIZE;
				if (file_data.size() < start_index + DATA_MSG_SIZE)
				{
					file_data.resize(start_index + DATA_MSG_SIZE);
				}
				
				std::copy(PreviousDataBuffer, PreviousDataBuffer + DATA_MSG_SIZE, file_data.begin() + start_index);
				log("Repeated DATA_MSG block " + std::to_string(current_block_number) + " with size " + std::to_string(DATA_MSG_SIZE));
				file_size_bytes_recd += DATA_MSG_SIZE;
				current_crc32 = previous_crc32;
				sendAckMessage(current_block_number, current_crc32);
				current_block_number++;
			}
		}
		else
		{
			// This is an unexpected message size - must be last block
			// However, if it is exactly SHORT_MSG_SIZE + 1, 
			// either the last remaining file data happened to be SHORT_MSG_SIZE + 1,
			// or it was actually SHORT_MSG_SIZE, and zero padded to make it SHORT_MSG_SIZE + 1
			// How can we know the difference? Check the expected file size against
			// what it would be with this last block...
			if (file_size_bytes_recd + recv_data.size() == file_size + 1)
			{
				// Remove the last byte from recv_data...
				recv_data.pop_back();
				// Re-calculate the CRC32
				current_crc32 = calculateCRC32((char*)DataBuffer, recv_data.size());
				log("Removed pad byte from last block");
			}



			// Copy the block of data into the DataBuffer

			uint32_t start_index = current_block_number * DATA_MSG_SIZE;
			if (file_data.size() < start_index + recv_data.size())
			{
				file_data.resize(start_index + recv_data.size());
			}
			std::copy(recv_data.begin(), recv_data.end(), file_data.begin() + start_index);
			log("Received DATA_MSG block " + std::to_string(current_block_number) + " with size " + std::to_string(recv_data.size()));
			file_size_bytes_recd += recv_data.size();
			sendAckMessage(current_block_number, current_crc32);
			current_block_number++;
		}
	}

    log("file_size_bytes expected: " + std::to_string(file_size));
    log("file_size_bytes_recd: " + std::to_string(file_size_bytes_recd));
    log("file_data.size(): " + std::to_string(file_data.size()));
	return false;
}


void Server::sendAckMessage(uint32_t blocknum, uint32_t crc32)
{
    BYTE tmpBuf[SHORT_MSG_SIZE];
    ZeroMemory(tmpBuf, SHORT_MSG_SIZE);

    // ACK_MSG is a CRC32 followed by a BLOCK_NUMBER
	uint32_t* crc32_ptr = reinterpret_cast<uint32_t*>(tmpBuf);
	uint32_t* blocknum_ptr = reinterpret_cast<uint32_t*>(tmpBuf + sizeof(uint32_t));
	*crc32_ptr = crc32;
	*blocknum_ptr = blocknum;

	std::vector<uint8_t> serializedMessage;
	serializedMessage.reserve(SHORT_MSG_SIZE);
	serializedMessage.assign(tmpBuf, tmpBuf + sizeof(tmpBuf));
	udpSend(serializedMessage);
	log("Sent ACK_MSG for block " + std::to_string(blocknum));
}