// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 25 May 2024
// License: This work is licensed under Creative Commons Attribution-ShareAlike 4.0 International 
// (CC BY-SA 4.0). https://creativecommons.org/licenses/by-sa/4.0/

#include "UDP_Receive.h"
#include <winsock2.h>
#include <ws2tcpip.h>

UPD_Receive::UPD_Receive(PCSTR address, u_short port)
{
    // Create a socket
    udpSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (udpSocket == INVALID_SOCKET) 
    {
        std::cerr << "Socket creation failed with error: " << WSAGetLastError() << std::endl;
        WSACleanup();
        return;
    }

    // Set up the sockaddr_in structure
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_port = htons(port);
    inet_pton(AF_INET, address, &serverAddr.sin_addr);

    if (bind(udpSocket, (SOCKADDR*)&serverAddr, sizeof(serverAddr)) == SOCKET_ERROR)
    {
        std::cerr << "bind failed with error: " << WSAGetLastError() << std::endl;
        closesocket(udpSocket);
        WSACleanup();
        return;
    }
}

UPD_Receive::~UPD_Receive()
{
    closesocket(udpSocket);
    // Free the allocated memory
    data.clear();
    WSACleanup();
}

std::vector<uint8_t> UPD_Receive::exampleUDPReceiveFunction(bool* pbSuccess)
{
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(udpSocket, &readfds);

    struct timeval timeout;
    timeout.tv_sec = TIMEOUT / 1000;
    timeout.tv_usec = (TIMEOUT % 1000) * 1000;

    int selectResult = select(0, &readfds, nullptr, nullptr, &timeout);
    if (selectResult > 0 && FD_ISSET(udpSocket, &readfds))
    {
        // Receive the data
        int serverAddrSize = sizeof(serverAddr);
        char buffer[512];
        memset(buffer, 0, sizeof(buffer));
        int result = recvfrom(udpSocket, buffer, 512, 0, reinterpret_cast<sockaddr*>(&serverAddr), &serverAddrSize);
        if (result == SOCKET_ERROR)
        {
            *pbSuccess = false;
            std::cerr << "Receive failed with error: " << WSAGetLastError() << std::endl;
        }
        else
        {
            *pbSuccess = true;
            // Copy the buffer to the data vector...
            data.resize(result);
            std::copy(buffer, buffer + result, data.begin());
        }
    }
    else
    {
        *pbSuccess = false;
        std::cerr << "Receive timed out." << std::endl;
    }

    return data;
}
