// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 25 May 2024
// License: This work is licensed under Creative Commons Attribution-ShareAlike 4.0 International 
// (CC BY-SA 4.0). https://creativecommons.org/licenses/by-sa/4.0/

#include <iostream>
#include "Client.h"

DWORD WINAPI RcvrThread(void* param)
{
    Client* pClient = (Client*)param;
    pClient->log("Thread started.");

    while (pClient->bThreadLive)
    {
        if (WaitForSingleObject(pClient->hEvent, MAX_TIMEOUT) == WAIT_OBJECT_0)
        {
            if (pClient->bThreadLive)
            {
                // Wait for acknowledgment
                pClient->bAckValid = false;
                pClient->bResend = false;
                bool bSuccess = false;
				auto recv_data = pClient->udpReceive(&bSuccess); // 53 usecs
                if (!bSuccess)
                {
                    pClient->log("udpReceive failed - socket error or time out!");
                }
                else
                {
                    ACK_RESULT result = pClient->VerifyMessage(recv_data, pClient->current_block_number, pClient->current_crc32, pClient->previous_crc32);
                    if (result == ACK_RESULT::ACK_OK)
                    {
                        pClient->bAckValid = true;
                    }
                    else if (result == ACK_RESULT::ACK_FOR_PREVIOUS_BLOCK)
                    {
                        pClient->bAckValid = true;
                        pClient->bResend = true;
                    }
                }

                pClient->bAckRecd = true; // We can exit the wait now
            }
        }
    }

	pClient->log("Thread terminated.");
    return 0;
}


bool Client::sendFile(const std::string& filename, LPBYTE& file_data, uint32_t& file_size)
{
    LPBYTE pFile = file_data;
    uint32_t file_bytes_sent = 0;
    log("Bytes to send: " + std::to_string(file_size));
	std::cout << "Sending file: " << filename << std::endl;
	std::cout << "Bytes to send: " << file_size << std::endl;

    // Construct the header message
    // HDR_MSG: a FILE_SIZE followed by a FILENAME for a total of HDR_MSG_SIZE bytes.
    // - HDR_MSG must not be SHORT_MSG_SIZE in total so there is no chance of the code mistaking it for a SHORT_MSG.
    // - this can be assured by simply appending a zero as necessary.

    ZeroMemory(DataBuffer, DATA_MSG_SIZE);

	HdrBuffer* pHdrBuffer = (HdrBuffer*)DataBuffer;
	pHdrBuffer->filesize = file_size;
    strcpy((char*)&DataBuffer[sizeof(uint32_t)], filename.c_str());

    // How many bytes in the DataBuffer are actually used?
    uint32_t hdr_msg_size = sizeof(uint32_t) + filename.size();

    // HDR_MSG must not be SHORT_MSG_SIZE
    // so we need to add a zero as necessary.
    if (hdr_msg_size == SHORT_MSG_SIZE)++hdr_msg_size;

    // We need the CRC32 to compare with the receiver
    current_crc32 = calculateCRC32((char*)DataBuffer, hdr_msg_size);

	log("Sending HDR_MSG of " + std::to_string(hdr_msg_size) + " bytes.");
	log("CRC32: " + std::to_string(current_crc32));

    
    serializedMessage.resize(hdr_msg_size);
    memcpy(serializedMessage.data(), DataBuffer, hdr_msg_size);
    udpSend(serializedMessage);
	std::cout << "Awaiting receiver" << std::endl;

	current_block_number = 0;

    hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    bAckRecd = false;
	bThreadLive = true;
    hThread = CreateThread(NULL, 0, RcvrThread, this, 0, NULL);
    if (hThread == NULL)
    {
        log("Failed to create thread. Error: " + std::to_string(GetLastError()));
        return false;
    }

    if (!AwaitAcknowlegment(serializedMessage))
    {
        CleanupThread();
        return false;
    }
	std::cout << "Receiver ready" << std::endl;
	std::cout << "Sending data";

    // Send data messages
	ZeroMemory(DataBuffer, DATA_MSG_SIZE);
    uint32_t max_blocks = (file_size + DATA_MSG_SIZE - 1) / DATA_MSG_SIZE;

    for (uint32_t block_num = 0; block_num < max_blocks; ++block_num)
    {
        memcpy(PreviousDataBuffer, DataBuffer, DATA_MSG_SIZE);

        // How many bytes remain?
        uint32_t bytes_remaining = file_size - file_bytes_sent;
        uint32_t bytes_to_send = bytes_remaining > DATA_MSG_SIZE ? DATA_MSG_SIZE : bytes_remaining;

		memcpy(DataBuffer, pFile, bytes_to_send);
        pFile += bytes_to_send;

        previous_crc32 = current_crc32;
        current_crc32 = calculateCRC32((char*)DataBuffer, bytes_to_send);

        if (block_num > 0 && current_crc32 == previous_crc32 && bytes_to_send == DATA_MSG_SIZE)
        {
            bool bSendNextBlock = true;

            // CRCs were the same - was the data the same?
            if (isSameData(DataBuffer, PreviousDataBuffer))
            {
                // Emit a REP_MSG, which is a CRC32 followed by a BLOCK_NUMBER
                BYTE tmpBuf[SHORT_MSG_SIZE];
                ZeroMemory(tmpBuf, SHORT_MSG_SIZE);
                uint32_t* pCRC = (uint32_t*)tmpBuf;
                *pCRC = current_crc32;
                uint32_t* pBlockNum = (uint32_t*)&tmpBuf[sizeof(uint32_t)];
                serializedMessage.resize(SHORT_MSG_SIZE);
                memcpy(serializedMessage.data(), tmpBuf, SHORT_MSG_SIZE);
                udpSend(serializedMessage);

                log("Sent REP_MSG for block " + std::to_string(block_num) + ", crc32: " + std::to_string(current_crc32));
                bSendNextBlock = false;
            }
            else
            {
                // CRC32 collision - not expected!
                // If the data is not identical, send a DUP_WARN_MSG,
                // then send the block itself as a DATA_MSG. 
                // Emit a DUP_MSG, which is a CRC32 followed by a DUP_WARN  
                BYTE tmpBuf[SHORT_MSG_SIZE];
                ZeroMemory(tmpBuf, SHORT_MSG_SIZE);
                uint32_t* pCRC = (uint32_t*)tmpBuf;
                *pCRC = current_crc32;
                uint32_t* pDupWarn = (uint32_t*)&tmpBuf[sizeof(uint32_t)];
                *pDupWarn = DUP_WARN;

                serializedMessage.resize(SHORT_MSG_SIZE);
                memcpy(serializedMessage.data(), tmpBuf, SHORT_MSG_SIZE);
                udpSend(serializedMessage);
                log("Sent DUP_WARN for block " + std::to_string(block_num) + ", crc32: " + std::to_string(current_crc32));
            }

            // Wait for acknowledgment of the REP_MSG or DUP_WARN message
			current_block_number = block_num;
            if (!AwaitAcknowlegment(serializedMessage))
            {
                CleanupThread();
                return false;
            }

			// If the ACK we received was for the previous block,
			// we need to resend the current block, because the receiver
			// didn't get it.
            if(bResend)
			{
				if (!HandleResends(serializedMessage))
                {
                    CleanupThread();
                    return false;
                }
            }
			file_bytes_sent += bytes_to_send;
			if (!bSendNextBlock)continue;
        }

        // Send the data message
		// The last block of the file may be less than 512 bytes
		// If it just happens to be SHORT_MSG_SIZE bytes, we need to pad it with zero
		// The receiver will know it was padded when it checks the file size.

        if (bytes_to_send == SHORT_MSG_SIZE)
        {
            ++bytes_to_send;
			log("Padded block " + std::to_string(block_num) + " with zero.");
        }
        serializedMessage.resize(bytes_to_send);
        memcpy(serializedMessage.data(), DataBuffer, bytes_to_send);
		udpSend(serializedMessage); // 56 usecs over the loopback
        file_bytes_sent += bytes_to_send;
		count_blocks_sent++;

        // Need some indication of progress
        if ((count_blocks_sent % 100) == 0)
        {
            std::cout << ".";
        }
        log("Sent DATA_MSG for block " + std::to_string(block_num));

		// Wait for acknowledgment of the data message
		current_block_number = block_num;
        
		if (!AwaitAcknowlegment(serializedMessage))
        {
            CleanupThread();
            return false;
        }
        if (bResend)
        {
            if (!HandleResends(serializedMessage))
            {
                CleanupThread();
                return false;
            }
        }
    }

    // When the SENDER completes sending all file data, it may display a success message to the user,
    // and sends a sequence of('MAX_RESENDS' + 1) EOF messages with a half - second pause between each and terminates,
    // without looking for a response from the RECEIVER.
	std::cout << std::endl;
	std::cout << "Sending EOFs";
    serializedMessage.resize(0);
    for (int i = 0; i <= MAX_RESENDS; i++)
    {
        // Send the EOF message of zero bytes
        udpSend(serializedMessage);
		Sleep(500);
		std::cout << ".";
	}

    log("file_bytes_sent: " + std::to_string(file_bytes_sent) + " file_size: " + std::to_string(file_size));    
    CleanupThread();
    return true;
}
 

ACK_RESULT Client::VerifyMessage(const std::vector<uint8_t>& data, uint32_t blocknum, uint32_t crc32, uint32_t previous_crc32)
{
    if (data.size() != SHORT_MSG_SIZE)
    {
        log("udpReceive failed - expected SHORT_MSG_SIZE  - data size: " + std::to_string(data.size()));
		return ACK_RESULT::ACK_INVALID;
    }
    ShortMessage msg;
    memcpy(&msg, data.data(), sizeof(ShortMessage));
    if (msg.block_number == blocknum && msg.crc32 == crc32)
    {
        // log("Received valid ACK.");
		return ACK_RESULT::ACK_OK;
    }

    if (msg.block_number == (blocknum - 1) && msg.crc32 == previous_crc32)
	{
		log("Received ACK for previous block.");
		return ACK_RESULT::ACK_FOR_PREVIOUS_BLOCK;
	}

    log("Received invalid ACK.");
	return ACK_RESULT::ACK_INVALID;
}


bool Client::isSameData(LPBYTE pSendBuf, LPBYTE pPrevBuf)
{
    int i;
    for (i = 0; i < DATA_MSG_SIZE; i++)
    {
        if (*pSendBuf++ != *pPrevBuf++)break;
    }
    return (i == DATA_MSG_SIZE);
}


bool Client::AwaitAcknowlegment(const std::vector<uint8_t>& serializedMessage)
{
    bAckRecd = false;
	bAckValid = false;

    // Fire the event to start the thread
	SetEvent(hEvent);

    int count = 0;
    while (!bAckRecd)
    {
        for (int i = 0; i < 10; i++)
        {
            Sleep(1);
            if (bAckRecd)
            {
				totalWaitMS += i;
                break;
            }
        }

        if (bAckRecd)break;
        totalWaitMS += 10;

        // The SENDER is broadcasting a HDR_MSG every 2 seconds until it gets an ACK_MSG
        // or until it has tried MAX_RESENDS times.
        log("Waiting for ACK... Attempt: " + std::to_string(count));

        for (int i = 0; i < 100; i++)
        {
            Sleep(20);
            if (bAckRecd)
            {
				totalWaitMS += (i * 20);
                break;
            }
        }

        if (bAckRecd)break;

        if (count++ > MAX_RESENDS)
        {
            log("Failed to receive ACK_MSG for HDR_MSG.");
            return false;
        }
        udpSend(serializedMessage);
    }
    if (!bAckValid)
    {
        log("Failed to receive valid ACK_MSG for HDR_MSG.");
        return false;
    }

    return true;
}

bool Client::HandleResends(const std::vector<uint8_t>& serializedMessage)
{
	int countResends = 0;
    while (bResend)
    {
        if (MAX_RESENDS == countResends)
        {
			log("Failed to receive ACK_MSG after MAX_RESENDS.");
			return false;
		}
		bResend = false;

		log("Resending block " + std::to_string(current_block_number));
		udpSend(serializedMessage);
		++countResends;

		// AwaitAcknowlegment will set bResend to true if the ACK is for the previous block
		if (!AwaitAcknowlegment(serializedMessage))return false;
	}
	return true;
}

void Client::CleanupThread(void)
{
	if (hThread == NULL)return;
	if (!bThreadLive)return;
	bThreadLive = false;
	SetEvent(hEvent);
	DWORD waitResult = WaitForSingleObject(hThread, INFINITE);
    switch (waitResult)
    {
	case WAIT_OBJECT_0:
		log("Thread terminated normally.");
		break;
	default:
		log("Thread wait failed. Error: " + std::to_string(GetLastError()));
		break;
	}
	CloseHandle(hThread);
	CloseHandle(hEvent);

    log("count_blocks_sent " + std::to_string(count_blocks_sent));
    log("Total wait time: " + std::to_string(totalWaitMS) + " ms");
}
