// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 25 May 2024
// License: This work is licensed under Creative Commons Attribution-ShareAlike 4.0 International 
// (CC BY-SA 4.0). https://creativecommons.org/licenses/by-sa/4.0/


#pragma once
#include <Windows.h> // only for LPBYTE
#include <cstdint>
#include <vector>
#include <string>
#include <functional>

// Define constants
constexpr uint32_t MAX_RESENDS = 10;            // Maximum number of resends
constexpr uint32_t DATA_MSG_SIZE = 512;         // Size of data message in bytes
constexpr uint32_t SHORT_MSG_SIZE = 8;          // Size of short message in bytes
constexpr uint32_t DUP_WARN = 0xFFFFFFFF;       // DUP_WARN value
constexpr uint32_t SOCKET_TIMEOUT = 20;         // Socket timeout in seconds

/*
From the protocol documantation:

FILE_SIZE: 32 bit unsigned value
CRC32 : 32 bit unsigned value
BLOCK_NUMBER : 32 bit unsigned value
SHORT_MSG_SIZE : 8 bytes
FILENAME : variable size - 1 to 255 bytes
HDR_MSG_SIZE : variable size - 5 to 259 bytes zero padded such that HDR_MSG_SIZE != SHORT_MSG_SIZE
DUP_WARN : 0xFFFFFFFF
WAIT_TIME 2 seconds
MAX_WAIT_TIMES : 10
MAX_RESENDS : 10
SOCKET_TIMEOUT : 20 seconds(adjust as required)


HDR_MSG : a FILE_SIZE followed by a FILENAME for a total of HDR_MSG_SIZE bytes.
* HDR_MSG must not be SHORT_MSG_SIZE in total so there is no chance of the code mistaking it for a SHORT_MSG.
* This can be assured by simply appending a zero as necessary.
* the first packet received must be assumed to be the one and only HDR_MSG.


SHORT_MSG : a CRC32 followed by a BLOCK_NUMBER for a total of SHORT_MSG_SIZE bytes
There are three types of SHORT_MSG...
1.   ACK_MSG, which is a CRC32 followed by a BLOCK_NUMBER
2.   REP_MSG, which is a CRC32 followed by a BLOCK_NUMBER
3.   DUP_MSG, which is a CRC32 followed by a DUP_WARN(optional)


Rather than having some kind of package identifier sent along with the packet,
The number of bytes received is a key fundamental part of the protocol.
The type of message received can only be identified by the number of bytes received.

Any message that is SHORT_MSG sized bytes must be one of these 3 types, but how do we know an ACK_MSG from a REP_MSG from a DUP_MSG when received ? The sender only ever receives an ACK message, so any SHORT_MSG sized message must be that.The receiver can only receive either a REP_MSG or a DUP_MSG, and can distinguish between them by looking at the BLOCK_NUMBER to see if it is a DUP_WARN.


DATA_MSG : 1 to 512 bytes
* Full 512 byte blocks of file data are sent until last block of file, which of course may be anywhere from 1 to 512 bytes.
* measures must be taken to ensure an 8 byte DATA_MSG cannot be mistaken for a SHORT_MSG by using an extra zero pad byte(read on below).

DUP_WARN_MSG : DUP_MSG followed by a DATA_MSG sent separately and sequentially, with no pause between them to wait for an ACK.
* then the sender will expect an ACK for the DATA_MSG.If that doesn't arrive both the DUP_MSG and the DATA_MSG must be resent.
(Since sequential CRC32 collisions never happens in practice this part of the protocol may be omitted or not fully implemented)

EOF_MSG: 0 bytes
*/

union HdrBuffer
{
    uint8_t buffer[512];
    uint32_t filesize;
};

struct ShortMessage
{
    uint32_t crc32;
    uint32_t block_number;
};

// CRC function type
using CRCFunction = std::function<uint32_t(char*&, uint32_t&)>;

// Logging function type
using LoggingFunction = std::function<void(const std::string&)>;

// UDP send function type
using UDPSendFunction = std::function<void(const std::vector<uint8_t>&)>;

// UDP receive function type
using UDPReceiveFunction = std::function<std::vector<uint8_t>(bool*)>;

// MFTP class
class MFTP
{
public:
    // Constructor
    MFTP(CRCFunction crcFunc, LoggingFunction logFunc, UDPSendFunction udpSendFunc, UDPReceiveFunction udpReceiveFunc);

    // Helper functions
    uint32_t calculateCRC32(char* data, uint32_t size);
    void log(const std::string& message);
    void udpSend(const std::vector<uint8_t>& data);
    std::vector<uint8_t> udpReceive(bool* pbSuccess);

    // Exponential backoff
    void applyExponentialBackoff(uint32_t& current_timeout, uint32_t& resend_attempts);

    // Functions
    CRCFunction crcFunction;
    LoggingFunction loggingFunction;
    UDPSendFunction udpSendFunction;
    UDPReceiveFunction udpReceiveFunction;

    BYTE DataBuffer[DATA_MSG_SIZE] = {};
    BYTE PreviousDataBuffer[DATA_MSG_SIZE] = {};
    uint32_t current_block_number = 0;
    uint32_t current_crc32 = 0;
    uint32_t previous_crc32 = 0;
};
