// Author: George W. Taylor
// http://www.tropicalcoder.com/
// Date: 25 May 2024
// License: This work is licensed under Creative Commons Attribution-ShareAlike 4.0 International 
// (CC BY-SA 4.0). https://creativecommons.org/licenses/by-sa/4.0/

// Sender.cpp : Implements the Minimalist File Transfer Protocol (MFTP) sender application.

#include <winsock2.h>
#include <ws2tcpip.h>
#include <Windows.h>
#include <conio.h> // For _getch() to wait for a key press
#include <vector>
#include <string>
#include <iostream>
#include "Client.h"
#include "CheckCRC.h"
#include "Debug.h"
#include "FileLoader.h"
#include "UDP_Send.h"
#include "UDP_Receive.h"


// Link with Ws2_32.lib
#pragma comment(lib, "Ws2_32.lib")

CheckCRC* pCRC;
DebugLogging* pDebug;
Client* pClient;
FileLoader* pFileLoader;
UDP_Send* pUDPSend; 
UPD_Receive* pUDPReceive;
PCSTR address = "127.0.0.1";
// These ports are reversed in the receiver of course
u_short SendPort = 9788;
u_short RcvrPort = 9789;

// Don't send the full path with the filename because the receiver will not have the same path
// unless you are testing over the loopback, in which case it will try to write
// to the same file you are sending and that isn't gonna end well.
//char* filename = (char*)"RandumTest9.bin";


/*
// CRC function type
using CRCFunction = std::function<uint32_t(const std::vector<uint8_t>&)>;

// Logging function type
using LoggingFunction = std::function<void(const std::string&)>;

// UDP send function type
using UDPSendFunction = std::function<void(const std::vector<uint8_t>&)>;

// UDP receive function type
using UDPReceiveFunction = std::function<std::vector<uint8_t>()>;

*/

// Create a wrapper for the CRC32 calculation
static uint32_t exampleCRCFunction(char* pBuf, uint32_t countBytes)
{
	if (pCRC)
	{
		return pCRC->getCRC32(pBuf, countBytes);
	}
	return 0;
}

// Create a wrapper for the logging function
static void exampleLoggingFunction(const std::string& message) {
	if (pDebug)
	{
		pDebug->DebugMessage((char*)message.c_str());
	}
}

static void exampleUDPSendFunction(const std::vector<uint8_t>& data)
{
	if (pUDPSend)
	{
		pUDPSend->exampleUDPSendFunction(data);
	}
}

static std::vector<uint8_t> exampleUDPReceiveFunction(bool* pbSuccess)
{
	if (pUDPReceive)
	{
		return pUDPReceive->exampleUDPReceiveFunction(pbSuccess);
	}
	return std::vector<uint8_t>();
}

static bool isValidFilename(const std::string& filename)
{
	// Basic validation: ensure the filename is not empty and does not contain illegal characters
	if (filename.empty())
	{
		return false;
	}
	const std::string illegalChars = "\\/:?\"<>|";
	for (char c : filename)
	{
		if (illegalChars.find(c) != std::string::npos)
		{
			return false;
		}
	}
	return true;
}


int main(int argc, char* argv[])
{
	if (argc < 2) 
	{
		std::cout << "Please supply the filename for file to send as the command line argument." << std::endl;
		std::cout << "Any key exits...";
		_getch(); // Wait for any key press
		return 1;
	}

	std::string filename = argv[1];

	if (!isValidFilename(filename))
	{
		std::cout << "Invalid filename. Please supply a valid filename." << std::endl;
		std::cout << "Any key exits...";
		_getch(); // Wait for any key press
		return 1;
	}

    pCRC = new CheckCRC();
	pDebug = new DebugLogging();
	pDebug->InitDebugFile();


	// Initialize Winsock
	WSADATA wsaData;
	int result = WSAStartup(MAKEWORD(2, 2), &wsaData);
	if (result != 0)
	{
		std::cerr << "WSAStartup failed with error: " << result << std::endl;
		return 1;
	}

	pUDPSend = new UDP_Send(address, SendPort);

	pUDPReceive = new UPD_Receive(address, RcvrPort);

	Client* pClient = new Client(exampleCRCFunction, exampleLoggingFunction, exampleUDPSendFunction, exampleUDPReceiveFunction);


	pFileLoader = new FileLoader((char*)filename.c_str());
	LPBYTE pFileData = pFileLoader->GetFilePointer();
	DWORD dwFileSizeBytes = pFileLoader->GetFileSizeBytes();

	if ((pFileData == NULL) || (dwFileSizeBytes == 0))
	{
		pDebug->DebugMessage("Failed to read file!");
		std::cout << "Failed to read file!" << std::endl;
		return 1;
	}

	// We have dwFileSizeBytes of data at LPBYTE pFileData
	// We can now send the file
	uint32_t file_size = (uint32_t)dwFileSizeBytes;
	if (pClient->sendFile(filename, pFileData, file_size))
	{
		std::cout << std::endl;
		std::cout << "File sent successfully!" << std::endl;
	}
	else
	{
		std::cout << "Failed to send file!" << std::endl;
	}

	pDebug->CloseDebugFile();
	delete pCRC;
	delete pDebug;
	delete pFileLoader;
	delete pUDPSend;
	delete pUDPReceive;
	delete pClient;
	return 0;
}
